import numpy as np
import casadi as ca
import spatial_casadi as sc
from scipy.spatial.transform import Rotation as R
from grasp_planning.solver.robot_model import RobotKinematicModel
from grasp_planning.cost.costs import SquaredAccCost
from grasp_planning.constraints.constraints import *

class GOMP():
    def __init__(self, n_waypoints, urdf, theta=np.pi/4, roll_obj_grasp = np.pi, root_link='world', end_link='tool_frame') -> None:
        # Init variables
        self.T_W_Grasp = None
        self.T_W_Obj = None
        self.T_Obj_Grasp = np.eye(4, dtype=float)
        
        # Kinematics
        self._robot_model = RobotKinematicModel(urdf, root_link, end_link)
        self.n_dofs = self._robot_model.n_dofs
        self.manipulation_frame_dim = 6
        self.x_dim = self.n_dofs
        self.n_waypoints = n_waypoints
        self.theta = theta
        self.roll_obj_grasp = roll_obj_grasp
        
        # Optimization
        self.x = ca.SX.sym("x", self.x_dim, self.n_waypoints)
        self.x_init = None
        
        self._objective = None
        self.l_joint_limits, self.u_joint_limits = None, None
        self.g_list = []
        
        self.param_ca_dict = dict()
        self._param_ca = None
        self._param_num = None


    def update_grasp_DOF(self, theta: float, degrees=True) -> None:
        """
        Object has to have z-axis aligned with the world frame
        """
        if degrees:
            self.theta = np.deg2rad(theta)
        else:
            self.theta = theta
        # Create static grasp's frame with respect to object's frame
        self.T_Obj_Grasp = np.eye(4, dtype=float)
        self.T_Obj_Grasp[:3,:3] = R.from_euler('xyz', [self.roll_obj_grasp, 0, 0], degrees=False).as_matrix()

        # Compute lower bound with the additional DoF
        T_Grasp_DoF = np.eye(4)
        T_Grasp_DoF[:3,:3] = R.from_euler('xyz', [0.0, -self.theta, 0], degrees=False).as_matrix()

        # Lower bound with respect to the world frame
        self.T_W_Grasp = self.T_W_Obj @ self.T_Obj_Grasp @ T_Grasp_DoF


    def update_object_pose(self, T_W_Obj: np.array) -> None:
        """
        Update object and grasp poses
        """
        self.T_W_Obj = T_W_Obj

    def set_init_guess(self, q):
        self.x_init = np.tile(q, self.n_waypoints)

    def set_boundary_conditions(self, q_start, q_end=None):
        if self.l_joint_limits is None or self.u_joint_limits is None:
            # Reset joint limits
            self.l_joint_limits = np.full((self.x_dim, self.n_waypoints), -100.0)
            self.u_joint_limits = np.full((self.x_dim, self.n_waypoints), 100.0)

            # get joint limits
            joint_limits = self._robot_model.get_joint_pos_limits()
            for t in range(self.n_waypoints):
                self.l_joint_limits[:self.n_dofs, t] = joint_limits[:,0]
                self.u_joint_limits[:self.n_dofs, t] = joint_limits[:,1]
 
        # init boundary
        self.l_joint_limits[:,0] = q_start
        self.u_joint_limits[:,0] = q_start

        if q_end != None:
            print("Final boundary condition is not implemented.")
    
    

    def setup_problem(self, verbose=False):
        # assert self.T_Obj_Grasp != None, "Grasp DOF is not set. Set additional degree of freedom around object."
        # assert self.T_W_Obj != None, "Object pose is not set."
        # 1. create decision variable
        x_ca_flatten = self.x.reshape((-1,1))
        
        # assert self.x_init != None, "Initial guess is missing. Set up one using `set_init_guess(q)`."

        g, self.g_lb, self.g_ub = ca.vertcat(), ca.vertcat(), ca.vertcat()
        for g_term in self.g_list:
            gt, g_lb, g_ub = g_term.get_constraint()
            g = ca.vertcat(g, gt)
            self.g_lb = ca.vertcat(self.g_lb, g_lb)
            self.g_ub = ca.vertcat(self.g_ub, g_ub)


        # Read Symbolic Paramaters
        for param in self.param_ca_dict:
            if self.param_ca_dict[param]["sym_param"] is not None:
                if self._param_ca is None:
                    self._param_ca = self.param_ca_dict[param]["sym_param"].reshape((-1,1))
                else:
                    self._param_ca = ca.vertcat(self._param_ca, self.param_ca_dict[param]["sym_param"].reshape((-1,1)))

        options = {}
        options["ipopt.acceptable_tol"] = 1e-3
        if not verbose:
            options["ipopt.print_level"] = 0
            options["print_time"] = 0

        self.solver = ca.nlpsol('solver', 'ipopt', {'x': x_ca_flatten, 'f': self.objective.eval_cost(x_ca_flatten), 'g': g, 'p': self._param_ca}, options)
    

        
    def solve(self):
        # Replace symbolic variables with numeric values
        for param in self.param_ca_dict:
            if self.param_ca_dict[param]["num_param"] is not None:
                if self._param_num is None:
                    self._param_num = self.param_ca_dict[param]["num_param"].reshape((-1,1), order='F')
                else:
                    self._param_num = ca.vertcat(self._param_num, self.param_ca_dict[param]["num_param"].reshape((-1,1), order='F'))

        result = self.solver(x0=self.x_init,
                             lbg=self.g_lb,
                             ubg=self.g_ub,
                             lbx=self.l_joint_limits.reshape((-1,1), order='F'),
                             ubx=self.u_joint_limits.reshape((-1,1), order='F'),
                             p=self._param_num)

        success_flag = self.solver.stats()["success"]
        success_msg = self.solver.stats()["return_status"]
        return  result['x'].reshape((self.n_dofs, self.n_waypoints)), success_flag


    def add_objective_function(self):
        # Define parameter sym variable
        name = "objective_param"
        self.param_ca_dict[name] =  {
            "sym_param" : None,
            "num_param" : None
            }
        # Create objective function
        self.objective = SquaredAccCost(self.n_waypoints, self.x_dim, manip_frame=False)
    

    def update_constraints_params(self, content_dict):
        for g_name, g_param in content_dict.items():
            self.param_ca_dict[g_name]["num_param"] = g_param

    def add_collision_constraint(self, name, waypoint_ID, child_link, r_link=0.5, r_obst=0.2, tolerance=0.01):
        # Define parameter sym variable
        self.param_ca_dict[name] =  {
            "waypoint_ID" : waypoint_ID,
            "sym_param" : ca.SX.sym(name, 3, 1),
            "num_param" : np.zeros((3,1)),
            "tolerance" : tolerance,
            "child_link" : child_link,
            "r_link" : r_link,
            "r_obst" : r_obst
            }
        self.g_list.append(CollisionConstraint(robot=self._robot_model,
                                                q_ca=self.x,
                                                waypoint_ID=waypoint_ID,
                                                param_obst_ca=self.param_ca_dict[name]["sym_param"],
                                                link_name=child_link,
                                                r_link= r_link,
                                                r_obst= r_obst,
                                                tolerance=tolerance))

    def add_grasp_pos_constraint(self, name, waypoint_ID, tolerance=0.0):
        # Define parameter sym variable
        self.param_ca_dict[name] =  {
            "waypoint_ID" : waypoint_ID,
            "sym_param" : ca.SX.sym(name, 4, 4),
            "num_param" : np.eye(4),
            "tolerance" : tolerance
            }
        # Create
        self.g_list.append(GraspPosConstraint(robot=self._robot_model, 
                                              q_ca=self.x, 
                                              waypoint_ID=waypoint_ID, 
                                              paramca_T_W_Grasp=self.param_ca_dict[name]["sym_param"],
                                              tolerance=tolerance))

    def add_grasp_rot_constraint(self, name, waypoint_ID, tolerance=0.1):
        # Define parameter sym variable
        self.param_ca_dict[name] =  {
            "waypoint_ID" : waypoint_ID,
            "sym_param" : ca.SX.sym(name, 4, 4),
            "num_param" : np.eye(4),
            "tolerance" : tolerance
            }
        self.g_list.append(GraspRotConstraint(robot=self._robot_model, 
                                              q_ca=self.x, 
                                              waypoint_ID=waypoint_ID, 
                                              paramca_T_W_Grasp=self.param_ca_dict[name]["sym_param"], 
                                              theta=self.theta, 
                                              tolerance=tolerance))


    def add_z_pos_constraint(self, name, waypoint_ID, z_height, tolerance=0.01):
        self.param_ca_dict[name] =  {
            "waypoint_ID" : waypoint_ID,
            "sym_param" : None,
            "num_param" : None,
            "z_height" : z_height,
            "tolerance" : tolerance
            }
        self.g_list.append(ZheightConstraint(robot=self._robot_model,
                                             q_ca=self.x,
                                             waypoint_ID=waypoint_ID, 
                                             z_height = z_height,
                                             tolerance=tolerance))


import numpy as np
import casadi as ca
import spatial_casadi as sc
from scipy.spatial.transform import Rotation as R
from forwardkinematics import GenericURDFFk

class RobotKinematicModel():
    def __init__(self, urdf_file, root_link, end_link):
        with open(urdf_file, "r") as file:
            self.urdf = file.read()

        self.robot_fk = GenericURDFFk(
                            self.urdf,
                            root_link = root_link,
                            end_links= end_link
                        )
        self.root_link = root_link
        self.end_link = end_link
        self.n_dofs = self.robot_fk.n()
        self.joint_names = []
        self.joint_pos_limits = np.zeros((self.n_dofs, 2))
        self._read_joint_pos_limits()

    def compute_fk_ca(self, q_ca, end_link=None):
        if end_link == None:
            return self.robot_fk.casadi(q_ca, self.end_link, position_only=False)
        else:
            return self.robot_fk.casadi(q_ca, child_link=end_link, position_only=False)


    def compute_fk_rpy_ca(self, q_ca):
        fk_ca = self.compute_fk_ca(q_ca)
        R_ca_rpy = sc.Rotation.from_matrix(fk_ca[:3,:3]).as_euler("xyz") #convert rotation part of FK into rpy
        return ca.vertcat(fk_ca[:3,3], R_ca_rpy)

    def eval_fk(self, q):
        return self.robot_fk.numpy(q, self.end_link, position_only=False)

    def _read_joint_pos_limits(self):
        joints = self.robot_fk.robot.get_joint_info(self.root_link, self.end_link )
        i = 0
        for joint in joints:
            if joint.type in ["revolute", "prismatic"]:
                self.joint_names.append(joint.name)
                # self.joint_pos_limits.append([joint.limit.lower, joint.limit.upper])
                self.joint_pos_limits[i] = [joint.limit.lower, joint.limit.upper]
                i += 1


    def get_joint_pos_limits(self):
        return self.joint_pos_limits

    def get_link_names(self):
        return self.robot_fk.robot.link_names()



from grasp_planning.constraints.collision_constraint import *
from grasp_planning.constraints.grasp_pos_constraint import *
from grasp_planning.constraints.grasp_rot_constraint import *
from grasp_planning.constraints.manipulation_constraint import *
from grasp_planning.constraints.position_constraint import *
from grasp_planning.constraints.orientation_constraint import *
from grasp_planning.constraints.z_height_constraint import *


import numpy as np
import casadi as ca
import spatial_casadi as sc
from grasp_planning.constraints.constraint_template import Constraint

# class GraspRotConstraint(Constraint):
#     def __init__(self, robot, q_ca, waypoint_ID, param_grasp_ca, theta=0.0, tolerance=0.1) -> None:
#         super().__init__() 
#         # q = q_ca[:n_dof, waypoint_ID] -> joint space of the robot (decision variable)
#         # manip_frame = q_ca[n_dof:, waypoint_ID] -> manipulation frame (decision variable)
#         # manip frame_pos - T_W_Grasp_pos = 0
#         self._robot = robot
#         self.theta = theta
#         self.tolerance = tolerance
#         self.param_grasp_ca = param_grasp_ca
        
#         R_W_EEF = self._robot.compute_fk_ca(q_ca[:,waypoint_ID])
#         R_G_EEF = self.param_grasp_ca[:3,:3].T @ R_W_EEF[:3,:3]
#         Rpy_G_EEF = sc.Rotation.from_matrix(R_G_EEF).as_euler("xyz")

#         self.g = ca.vertcat(Rpy_G_EEF[0], Rpy_G_EEF[1], Rpy_G_EEF[2])
#         self.g_lb = ca.vertcat(-self.tolerance, -self.theta, -self.tolerance)
#         self.g_ub = ca.vertcat(self.tolerance, self.theta, self.tolerance)

#         self.g_eval = ca.Function('g_grasp_rot', [q_ca, param_grasp_ca], [self.g])

#     def get_constraint(self):
#         return self.g, self.g_lb, self.g_ub

#     def do_eval(self, q, T_W_Grasp):
#         return self.g_eval(q, T_W_Grasp)

class GraspRotConstraint(Constraint):
    def __init__(self, robot, q_ca, waypoint_ID, paramca_T_W_Grasp, theta=0.0, tolerance=0.1) -> None:
        super().__init__() 
        # q = q_ca[:n_dof, waypoint_ID] -> joint space of the robot (decision variable)
        # manip_frame = q_ca[n_dof:, waypoint_ID] -> manipulation frame (decision variable)
        # manip frame_pos - T_W_Grasp_pos = 0
        self._robot = robot
        self.theta = theta
        self.tolerance = tolerance
        
        R_W_EEF = self._robot.compute_fk_ca(q_ca[:,waypoint_ID])
        R_G_EEF = paramca_T_W_Grasp[:3,:3].T @ R_W_EEF[:3,:3]
        Rpy_G_EEF = sc.Rotation.from_matrix(R_G_EEF).as_euler("xyz")

        self.g = ca.vertcat(Rpy_G_EEF[0], Rpy_G_EEF[1], Rpy_G_EEF[2])
        self.g_lb = ca.vertcat(-self.tolerance, -self.tolerance, -self.tolerance)
        self.g_ub = ca.vertcat(self.tolerance, 2*self.theta, self.tolerance)

        self.g_eval = ca.Function('g_grasp_rot', [q_ca, paramca_T_W_Grasp], [self.g])

    def get_constraint(self):
        return self.g, self.g_lb, self.g_ub

    def do_eval(self, q, T_W_Grasp):
        return self.g_eval(q, T_W_Grasp)



import numpy as np
import casadi as ca
import spatial_casadi as sc
from grasp_planning.constraints.constraint_template import Constraint

class CollisionConstraint(Constraint):
    def __init__(self, robot, q_ca, waypoint_ID, param_obst_ca, link_name, r_link, r_obst, tolerance=0.0) -> None:
        super().__init__() 
        self._robot = robot
        self.tolerance = tolerance
        self.r_link = r_link
        self.r_obst = r_obst
        self.d_safety = tolerance
        
        T_W_link = self._robot.compute_fk_ca(q_ca[:,waypoint_ID], link_name)

        self.g = ca.norm_2(T_W_link[:3,3]-param_obst_ca) #obstacle_pos
        self.g_lb = self.r_link + self.r_obst + self.d_safety
        self.g_ub = float("inf")


    def get_constraint(self):
        return self.g, self.g_lb, self.g_ub

    def do_eval(self, q, T_W_Grasp):
        pass
        #return self.g_eval(q, T_W_Grasp)



import numpy as np
import casadi as ca
import spatial_casadi as sc
from grasp_planning.constraints.constraint_template import Constraint


class ManipulationConstraint(Constraint):
    def __init__(self, robot, q_ca, waypoint_ID, n_dofs) -> None:
        super().__init__() 
        # q = q_ca[:n_dof, waypoint_ID] -> joint space of the robot (decision variable)
        # manip_frame = q_ca[n_dof:, waypoint_ID] -> manipulation frame (decision variable)
        # FK(q)_rpy - manip frame = 0
        self._robot = robot
        
        self.g = self._robot.compute_fk_rpy_ca(q_ca[:n_dofs, waypoint_ID]) - q_ca[n_dofs:, waypoint_ID] 
        self.g_lb = np.zeros(6)
        self.g_ub = np.zeros(6)
        self.g_eval = ca.Function('g_manip', [q_ca], [self.g])

        
    def get_constraint(self):
        return self.g, self.g_lb, self.g_ub
    
    def get_limits(self):
        return self.g_lb, self.g_ub
    
    def set_limits(self, lb, ub):
        self.g_lb = lb
        self.g_ub = ub
    
    def do_eval(self, q):
        return self.g_eval(q)



import numpy as np
import casadi as ca
import spatial_casadi as sc
from scipy.spatial.transform import Rotation as R
from grasp_planning.solver.robot_model import RobotKinematicModel
from grasp_planning.cost.costs import *
from grasp_planning.constraints.constraints import *

class IK_OPTIM():
    def __init__(self, urdf, root_link='world', end_link='tool_frame') -> None:
        # Init variables
        self.T_W_Ref = None
        
        # Kinematics
        self._robot_model = RobotKinematicModel(urdf, root_link, end_link)
        self.n_dofs = self._robot_model.n_dofs
        self.x_dim = self.n_dofs

        # Optimization
        self.x = ca.SX.sym("x", self.x_dim, 1)
        self.x_init = None
        
        self._objective = None
        self.l_joint_limits, self.u_joint_limits = None, None
        self.g_list = []

        self.param_ca_dict = dict()


    def set_init_guess(self, q):
        self.x_init = q

    def set_boundary_conditions(self):
        if self.l_joint_limits is None or self.u_joint_limits is None:
            # Reset joint limits
            self.l_joint_limits = np.full((self.x_dim, 1), -100.0)
            self.u_joint_limits = np.full((self.x_dim, 1), 100.0)

            # get joint limits
            joint_limits = self._robot_model.get_joint_pos_limits()
            self.l_joint_limits[:, 0] = joint_limits[:,0]
            self.u_joint_limits[:, 0] = joint_limits[:,1]

    

    def setup_problem(self, verbose=False):
        # Read Constraints
        g, self.g_lb, self.g_ub = ca.vertcat(), ca.vertcat(), ca.vertcat()
        for g_term in self.g_list:
            gt, g_lb, g_ub = g_term.get_constraint()
            g = ca.vertcat(g, gt)
            self.g_lb = ca.vertcat(self.g_lb, g_lb)
            self.g_ub = ca.vertcat(self.g_ub, g_ub)


        # Read Symbolic Paramaters
        for i, param in enumerate(self.param_ca_dict):
            if i == 0:
                self._param_ca = self.param_ca_dict[param]["sym_param"]
            else:
                self._param_ca = ca.vertcat(self._param_ca, self.param_ca_dict[param]["sym_param"].reshape((-1,1)))

        # Solver settings
        options = {}
        options["ipopt.acceptable_tol"] = 1e-3
        if not verbose:
            options["ipopt.print_level"] = 0
            options["print_time"] = 0

        # Define solver
        self.solver = ca.nlpsol('solver', 'ipopt', {'x': self.x, 'f': self.objective.eval_cost(self.x), 'g': g, 'p': self._param_ca}, options)
    

        
    def solve(self):
        # Replace symbolic variables with numeric values
        for i, param in enumerate(self.param_ca_dict):
            if i == 0:
                self._param_num = self.param_ca_dict[param]["num_param"]
            else:
                self._param_num = ca.vertcat(self._param_num, self.param_ca_dict[param]["num_param"].reshape((-1,1), order='F'))

        result = self.solver(x0=self.x_init,
                             lbg=self.g_lb,
                             ubg=self.g_ub,
                             lbx=self.l_joint_limits.reshape((-1,1), order='F'),
                             ubx=self.u_joint_limits.reshape((-1,1), order='F'),
                             p=self._param_num)

        success_flag = self.solver.stats()["success"]
        success_msg = self.solver.stats()["return_status"]
        return  result['x'], success_flag
    

    def add_objective_function(self, name):
        # Define parameter sym variable
        self.param_ca_dict[name] =  {
            "sym_param" : ca.SX.sym(name, self.x_dim, 1),
            "num_param" : np.zeros((self.x_dim, 1))
            }
        # Create objective function
        self.objective = DistToHome(q_home=self.param_ca_dict[name]["sym_param"] ,
                                    n_dofs=self.x_dim)
      
    def add_position_constraint(self, name, tolerance=0.0):
        # Define parameter sym variable
        self.param_ca_dict[name] =  {
            "sym_param" : ca.SX.sym(name, 4, 4),
            "num_param" : np.eye(4)
            }
        # Create constraint
        self.g_list.append(PositionConstraint(robot=self._robot_model,
                                              q_ca=self.x,
                                              paramca_T_W_Ref=self.param_ca_dict[name]["sym_param"],
                                              tolerance=tolerance))


    def add_orientation_constraint(self, name, tolerance=0.0):
        # Define parameter sym variable
        self.param_ca_dict[name] =  {
            "sym_param" : ca.SX.sym(name, 4, 4),
            "num_param" : np.eye(4),
            "tolerance" : tolerance
            }
        # Create constraint
        self.g_list.append(OrientationConstraint(robot=self._robot_model,
                                                 q_ca=self.x,
                                                 paramca_T_W_Ref=self.param_ca_dict[name]["sym_param"],
                                                 tolerance=tolerance))



    

    def add_collision_constraint(self, name, link_names, r_link=0.5, r_obst=0.2, tolerance=0.01):
        # Define parameter sym variable
        self.param_ca_dict[name] =  {
            "sym_param" : ca.SX.sym(name, 3, 1),
            "num_param" : np.zeros((3,1)),
            "tolerance" : tolerance
            }
        # Create constraint
        for link in link_names:
            self.g_list.append(CollisionConstraint(robot=self._robot_model,
                                                    q_ca=self.x,
                                                    waypoint_ID=0,
                                                    param_obst_ca=self.param_ca_dict[name]["sym_param"],
                                                    link_name=link,
                                                    r_link= r_link,
                                                    r_obst= r_obst,
                                                    tolerance=tolerance))




class Constraint():
    def __init__(self):
        pass

    def get_constraint(self):
        print("get_constraint method is not implemented.")

    def do_eval(self, q):
        print("do_eval method is not implemented.")



import numpy as np
import casadi as ca
import spatial_casadi as sc
from grasp_planning.constraints.constraint_template import Constraint

class GraspPosConstraint(Constraint):
    def __init__(self, robot, q_ca, waypoint_ID, paramca_T_W_Grasp, tolerance=0.0) -> None:
        super().__init__() 
        # q = q_ca[:n_dof, waypoint_ID] -> joint space of the robot (decision variable)
        # manip_frame = q_ca[n_dof:, waypoint_ID] -> manipulation frame (decision variable)
        # manip frame_pos - T_W_Grasp_pos = 0
        self._robot = robot
        self.tolerance = tolerance
        Rpy_W_EEF = self._robot.compute_fk_rpy_ca(q_ca[:,waypoint_ID])
        
        self.g = Rpy_W_EEF[:3] - paramca_T_W_Grasp[:3,3]
        self.g_lb = np.zeros(3) - self.tolerance
        self.g_ub = np.zeros(3) + self.tolerance
        self.g_eval = ca.Function('g_grasp_pos', [q_ca, paramca_T_W_Grasp], [self.g])
      
    def get_constraint(self):
        return self.g, self.g_lb, self.g_ub

    def do_eval(self, q, T_W_Grasp):
        return self.g_eval(q, T_W_Grasp)



import numpy as np
import casadi as ca
import spatial_casadi as sc
from grasp_planning.constraints.constraint_template import Constraint

class ZheightConstraint(Constraint):
    def __init__(self, robot, q_ca, waypoint_ID, z_height, tolerance=0.0) -> None:
        super().__init__() 
        self._robot = robot
        self.tolerance = tolerance
        T_W_EEF = self._robot.compute_fk_ca(q_ca[:,waypoint_ID])
        
        self.g = T_W_EEF[2,3]
        self.g_lb = z_height - self.tolerance
        self.g_ub = float("inf")
        self.g_eval = ca.Function('g_z_coord', [q_ca], [self.g])
      
    def get_constraint(self):
        return self.g, self.g_lb, self.g_ub

    def do_eval(self, q):
        return self.g_eval(q)



import numpy as np

"""
Finite Differencing Matrix to produce sum of squared accelerations
ddq= A * q
ddq.T * ddq = ddq.T * (A.T * A) * ddq
ddq.T * ddq = ddq.T * P * ddq
"""

class DistToHome():
    def __init__(self, q_home, n_dofs):
        self._n_dofs = n_dofs 
        self.q_home = q_home
        self.I = np.eye(self._n_dofs)

    def eval_cost(self, q):
        return (q -self.q_home).T @ self.I @ (q - self.q_home)